@@ -44,13 +44,16 @@ log = logging.getLogger("multissl")
 
 OPENSSL_OLD_VERSIONS = [
     "1.1.1w",
+    "3.1.8",
 ]
 
 OPENSSL_RECENT_VERSIONS = [
     "3.0.18",
-    "3.1.7",
     "3.2.6",
     "3.3.5",
+    "3.4.3",
+    "3.5.4",
+    # See make_ssl_data.py for notes on adding a new version.
 ]
 
 LIBRESSL_OLD_VERSIONS = [
@@ -68,9 +71,8 @@ MULTISSL_DIR = os.path.abspath(os.path.join(PYTHONROOT, '..', 'multissl'))
 parser = argparse.ArgumentParser(
     prog='multissl',
     description=(
-        "Run CPython tests with multiple OpenSSL and LibreSSL "
-        "versions."
-    )
+        "Run CPython tests with multiple cryptography libraries/versions."
+    ),
 )
 parser.add_argument(
     '--debug',
